/*
 *  Copyright (C) 2007 Badri Krishnan
 *  Copyright (C) 2008 Christine Chung, Badri Krishnan and John Whelan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with with program; see the file COPYING. If not, write to the 
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *  MA  02111-1307  USA
 */

/**
 * \author Christine Chung, Badri Krishnan, John Whelan
 * \date 2008
 * \file 
 * \ingroup pulsar
 * \brief Header-file for LAL routines for CW cross-correlation searches
 *
 * $Id: PulsarCrossCorr.h,v 1.2 2008/10/22 14:00:11 whelan Exp $
 *
 */
 
/*
 *   Protection against double inclusion (include-loop protection)
 *     Note the naming convention!
 */

#ifndef _PULSARCROSSCORR_H
#define _PULSARCROSSCORR_H

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glob.h>
#include <time.h>
#include <errno.h> 

#include <lal/AVFactories.h>
#include <lal/Date.h>
#include <lal/DetectorSite.h>
#include <lal/LALDatatypes.h>
#include <lal/LALHough.h>
#include <lal/RngMedBias.h>
#include <lal/LALRunningMedian.h>
#include <lal/Velocity.h>
#include <lal/Statistics.h>
#include <lal/ComputeFstat.h>
#include <lal/UserInput.h>
#include <lal/SFTfileIO.h>
#include <lal/NormalizeSFTRngMed.h>
#include <lal/LALInitBarycenter.h>
#include <lal/SFTClean.h>
#include <gsl/gsl_cdf.h>
#include <lal/FrequencySeries.h>
#include <lal/Sequence.h>


/******************************************************
 *   Protection against C++ name mangling
 */

#ifdef  __cplusplus
extern "C" {
#endif


/******************************************************
 *  Assignment of Id string using NRCSID()
 */

NRCSID (PULSARCROSSCORRH, "$Id: PulsarCrossCorr.h,v 1.2 2008/10/22 14:00:11 whelan Exp $");

/******************************************************
 *  Error codes and messages.
 */
 
#define PULSARCROSSCORR_ENULL 1
#define PULSARCROSSCORR_ENONULL 2

#define PULSARCROSSCORR_MSGENULL "Null pointer"
#define PULSARCROSSCORR_MSGENONULL "Non-null pointer"

/* ******************************************************************
 *  Structure, enum, union, etc., typdefs.
 */

  /** struct holding info about skypoints */
  typedef struct tagSkyPatchesInfo{
    UINT4 numSkyPatches;
    REAL8 *alpha;
    REAL8 *delta;
    REAL8 *alphaSize;
    REAL8 *deltaSize;
  } SkyPatchesInfo;

  /** A pair of SFTs which will be correlated 
      -- this struct contains all information needed 
      to calculate the cross correlation for some given 
      parameter space point   */
  typedef struct tagSingleSFTpair{
    COMPLEX8FrequencySeries  *sft1; 
    COMPLEX8FrequencySeries  *sft2; 
    REAL8FrequencySeries     *psd1;
    REAL8FrequencySeries     *psd2;
    REAL8 vel1[3];
    REAL8 vel2[3];
    REAL8 pos1[3];
    REAL8 pos2[3];
  } SingleSFTpair;

  /** vector of SFT pairs */
  typedef struct tagSFTPairVec{
    UINT4 length;
    SingleSFTpair *data;
  } SFTPairVec;  

  typedef struct tagSFTPairparams{
    REAL8 lag;
  } SFTPairParams;

  typedef struct tagSFTDetectorInfo{
    COMPLEX8FrequencySeries *sft;
    REAL8 vDetector[3];
    REAL8 rDetector[3];
    REAL8 a;
    REAL8 b;
  } SFTDetectorInfo;

/*
 *  Functions Declarations (i.e., prototypes).
 */

void LALCombineAllSFTs ( LALStatus *status,
			 SFTVector **outsfts,
			 MultiSFTVector *multiSFTs,
			 REAL8 length);

void LALCreateSFTPairsIndicesFrom2SFTvectors(LALStatus          *status,
					     INT4VectorSequence **out,
					     SFTVector          *in,
					     SFTPairParams      *par,
					     INT4 		detChoice);

void LALCorrelateSingleSFTPair(LALStatus                *status,
			       COMPLEX16                *out,
			       COMPLEX8FrequencySeries  *sft1,
			       COMPLEX8FrequencySeries  *sft2,
			       REAL8FrequencySeries     *psd1,
			       REAL8FrequencySeries     *psd2,
			       REAL8                    *freq1,
			       REAL8                    *freq2);

void LALGetSignalFrequencyInSFT(LALStatus                *status,
				REAL8                    *out,
				COMPLEX8FrequencySeries  *sft1,
				PulsarDopplerParams      *dopp,
				REAL8Vector              *vel,
				LIGOTimeGPS	         *firstTimeStamp);

void LALGetSignalPhaseInSFT(LALStatus               *status,
			    REAL8                   *out,
			    COMPLEX8FrequencySeries *sft1,
			    PulsarDopplerParams     *dopp,
			    REAL8Vector             *pos);

void LALCalculateSigmaAlphaSq(LALStatus            *status,
			      REAL8                *out,
			      REAL8                freq1,
			      REAL8                freq2,
			      REAL8FrequencySeries *psd1,
			      REAL8FrequencySeries *psd2);

void LALCalculateUalpha(LALStatus *status,
			COMPLEX16 *out,
			REAL8     *Aplus,
			REAL8     *Across,
			REAL8     *phiI,
			REAL8     *phiJ,
			REAL8     *FplusI,
			REAL8     *FplusJ,
			REAL8     *FcrossI,
			REAL8     *FcrossJ,
			REAL8     *sigmasq);

void LALCalculateCrossCorrPower(LALStatus       *status,
				REAL8	        *out,
				COMPLEX16Vector *yalpha,
				COMPLEX16Vector *ualpha);

void LALNormaliseCrossCorrPower(LALStatus        *status,
				REAL8		 *out,
				COMPLEX16Vector  *ualpha,
				REAL8Vector      *sigmaAlphasq);

/* ****************************************************** */

#ifdef  __cplusplus
}                /* Close C++ protection */
#endif


#endif     /* Close double-include protection _PULSARCROSSCORR_H */
